// ---------------------------------------------------------------------------
//  Includes for all the program files to see
// ---------------------------------------------------------------------------
#include <string.h>
#include <stdlib.h>
#include <iostream>
using namespace std;
#include <xercesc/util/PlatformUtils.hpp>

// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------
#include <xercesc/framework/StdInInputSource.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <xercesc/util/OutOfMemoryException.hpp>

// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------
#include <xercesc/sax/AttributeList.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/sax/SAXException.hpp>

#include "ViewCellsParser.h"

#include "ViewCellsParserXerces.h"
#include "Mesh.h"
#include "VspBspTree.h"
#include "ViewCellBsp.h"
#include "ViewCellsManager.h"

// ---------------------------------------------------------------------------
//  Local data
//
//  doNamespaces
//      Indicates whether namespace processing should be enabled or not.
//      The default is no, but -n overrides that.
//
//  doSchema
//      Indicates whether schema processing should be enabled or not.
//      The default is no, but -s overrides that.
//
//  schemaFullChecking
//      Indicates whether full schema constraint checking should be enabled or not.
//      The default is no, but -s overrides that.
//
//  valScheme
//      Indicates what validation scheme to use. It defaults to 'auto', but
//      can be set via the -v= command.
// ---------------------------------------------------------------------------
static bool     doNamespaces       = false;
static bool     doSchema           = false;
static bool     schemaFullChecking = false;
static SAXParser::ValSchemes    valScheme       = SAXParser::Val_Auto;





// ---------------------------------------------------------------------------
//  StdInParseHandlers: Constructors and Destructor
// ---------------------------------------------------------------------------
ViewCellsParseHandlers::ViewCellsParseHandlers(VspBspTree *tree,
											   ViewCellsManager *viewCells,
											   ObjectContainer *objects):
  mElementCount(0)
  , mAttrCount(0)
  , mCharacterCount(0)
  , mSpaceCount(0)
{
	mVspBspTree = tree;
	mCurrentNode = tree->GetRoot();
	mViewCellsManager = viewCells;
	mObjects = objects;
}

ViewCellsParseHandlers::~ViewCellsParseHandlers()
{
}


// ---------------------------------------------------------------------------
//  StdInParseHandlers: Implementation of the SAX DocumentHandler interface
// ---------------------------------------------------------------------------


void ViewCellsParseHandlers::endElement(const XMLCh* const name)
{
  StrX lname(name);
  string element(lname.LocalForm());
  if (element == "Interior")
	  EndInterior();
  if (element == "ViewCells")
	  EndViewCells();
}


void ViewCellsParseHandlers::EndInterior()
{
	// go one up in the tree
	if (mCurrentNode->GetParent())
	{	cout << "]";
		mCurrentNode = mCurrentNode->GetParent();
	}
}


inline bool vlt(ViewCell *v1, ViewCell *v2)
{
	return v1->mId < v2->mId;
}


void ViewCellsParseHandlers::EndViewCells()
{
	// sort view cells for easily finding id
	stable_sort(mViewCells.begin(), mViewCells.end(), vlt);
}


void ViewCellsParseHandlers::startElement(const XMLCh* const name,
										  AttributeList& attributes)
{
	StrX lname(name);
	string element(lname.LocalForm());
	
	if (element == "ViewCell")
	{
		cout << "v";
		StartViewCell(attributes);
	}

	if (element == "Interior") 
	{
		cout << "[";
		StartInterior(attributes);
	}

	if (element == "Leaf") 
	{
		StartLeaf(attributes);
	}

	++ mElementCount;
	mAttrCount += attributes.getLength();
}


inline bool ilt(Intersectable *obj1, Intersectable *obj2)
{
	return obj1->mId < obj2->mId;
}


void ViewCellsParseHandlers::StartViewCell(AttributeList&  attributes)
{
	int len = attributes.getLength();
	vector<int> objIndices;
  
	ViewCell *viewCell = mViewCellsManager->GenerateViewCell();
	mViewCells.push_back(viewCell);

	for (int i = 0; i < len; ++ i) 
	{
		string attrName(StrX(attributes.getName(i)).LocalForm());
		
		if (attrName == "pvs") 
		{
			StrX attrValue(attributes.getValue(i));
			
			// handle coordIndex
			objIndices.clear();
			const char *ptr = attrValue.LocalForm();
			char *endptr;
			
			while (1)
			{
				int index = strtol(ptr, &endptr, 10);

				if (ptr == endptr)
					break;

				objIndices.push_back(index);

				ptr = endptr;
			}

			//TODO: find objects and add them to pvs
			// TODO: get view cell with specified id
			MeshInstance dummyInst(NULL);

			vector<int>::const_iterator it, it_end = objIndices.end();
			for (it = objIndices.begin(); it != it_end; ++ it)
			{
				const int objId = *it;	
				dummyInst.SetId(objId);

				ObjectContainer::iterator oit =
					lower_bound(mObjects->begin(), mObjects->end(), &dummyInst, ilt);
			
				Intersectable *obj = *oit;
				
				if (obj->GetId() == objId)
				{
					viewCell->GetPvs().AddSample(obj);
				}
				else
				{
					Debug << "error: object does not exist" << endl;
				}
			}
		}
		else if (attrName == "id") 
		{
			StrX attrValue(attributes.getValue(i));
			
			const char *ptr = attrValue.LocalForm();
			char *endptr = NULL;
			const int id = strtol(ptr, &endptr, 10);

			viewCell->SetId(id);
		}
	}
}


void ViewCellsParseHandlers::StartLeaf(AttributeList& attributes)
{
	BspLeaf * leaf = 
		new BspLeaf(dynamic_cast<BspInterior *>(mCurrentNode), NULL);

	if (mCurrentNode) // replace front or (if not NULL) back child
	{
		dynamic_cast<BspInterior *>(mCurrentNode)->ReplaceChildLink(NULL, leaf);
	}
	else
	{
		mVspBspTree->mRoot = leaf;
	}

	//-- find associated view cell
	int viewCellId;
	
	int len = attributes.getLength();
	  
	for (int i = 0; i < len; ++ i) 
	{
		string attrName(StrX(attributes.getName(i)).LocalForm());
		StrX attrValue(attributes.getValue(i));

		const char *ptr = attrValue.LocalForm();
		char *endptr = NULL;

		if (attrName == "viewCellId") 
		{
			viewCellId = strtol(ptr, &endptr, 10);
		}
	}

	if (viewCellId >= 0) // valid view cell
	{
		// TODO: get view cell with specified id
		ViewCell dummyVc;
		dummyVc.SetId(viewCellId);

		ViewCellContainer::iterator vit =
			lower_bound(mViewCells.begin(), mViewCells.end(), &dummyVc, vlt);
			
		BspViewCell *viewCell = dynamic_cast<BspViewCell *>(*vit);
		if (viewCell->GetId() == viewCellId)
		{
			leaf->SetViewCell(viewCell);
		}
		else
		{
			Debug << "error: view cell does not exist" << endl;
		}
	}
	else
	{
		leaf->SetViewCell(mVspBspTree->GetOrCreateOutOfBoundsCell());
		leaf->SetTreeValid(false);
		mVspBspTree->PropagateUpValidity(leaf);
	}
}


void ViewCellsParseHandlers::StartInterior(AttributeList& attributes)
{
	Plane3 plane;
	int len = attributes.getLength();

	for (int i = 0; i < len; ++ i) 
	{
		string attrName(StrX(attributes.getName(i)).LocalForm());
		StrX attrValue(attributes.getValue(i));
		const char *ptr = attrValue.LocalForm();

		if (attrName == "plane") 
		{
			sscanf(ptr, "%f %f %f %f",
				   &plane.mNormal.x, &plane.mNormal.y, &plane.mNormal.z, &plane.mD);
		}
	}

	BspInterior* interior = new BspInterior(plane);
	
	if (mCurrentNode) // replace NULL child of parent with current node
	{
		BspInterior *current = dynamic_cast<BspInterior *>(mCurrentNode);

		current->ReplaceChildLink(NULL, interior);
		interior->SetParent(current);
	}
	else
	{
		mVspBspTree->mRoot = interior;
	}

	mCurrentNode = interior;
}


void ViewCellsParseHandlers::characters(const XMLCh* const chars,
										const unsigned int length)
{
	mCharacterCount += length;
}


void ViewCellsParseHandlers::ignorableWhitespace(const XMLCh* const chars,
												 const unsigned int length)
{
	mSpaceCount += length;
}


void ViewCellsParseHandlers::resetDocument()
{
	mAttrCount = 0;
	mCharacterCount = 0;
	mElementCount = 0;
	mSpaceCount = 0;
}


// ---------------------------------------------------------------------------
//  StdInParseHandlers: Overrides of the SAX ErrorHandler interface
// ---------------------------------------------------------------------------
void
ViewCellsParseHandlers::error(const SAXParseException& e)
{
  XERCES_STD_QUALIFIER cerr << "\nError at (file " << StrX(e.getSystemId())
			    << ", line " << e.getLineNumber()
			    << ", char " << e.getColumnNumber()
			    << "): " << StrX(e.getMessage()) << XERCES_STD_QUALIFIER endl;
}

void
ViewCellsParseHandlers::fatalError(const SAXParseException& e)
{
  XERCES_STD_QUALIFIER cerr << "\nFatal Error at (file " << StrX(e.getSystemId())
			    << ", line " << e.getLineNumber()
			    << ", char " << e.getColumnNumber()
			    << "): " << StrX(e.getMessage()) << XERCES_STD_QUALIFIER endl;
}

void
ViewCellsParseHandlers::warning(const SAXParseException& e)
{
  XERCES_STD_QUALIFIER cerr << "\nWarning at (file " << StrX(e.getSystemId())
			    << ", line " << e.getLineNumber()
			    << ", char " << e.getColumnNumber()
			    << "): " << StrX(e.getMessage()) << XERCES_STD_QUALIFIER endl;
}


bool ViewCellsParser::ParseFile(const string filename, 
								VspBspTree *tree,
								ViewCellsManager *viewCells,
								ObjectContainer *objects)
{
  // Initialize the XML4C system
  try {
    XMLPlatformUtils::Initialize();
  }
  
  catch (const XMLException& toCatch)
    {
      XERCES_STD_QUALIFIER cerr << "Error during initialization! Message:\n"
				<< StrX(toCatch.getMessage()) << XERCES_STD_QUALIFIER endl;
      return false;
    }
  
  
  //
  //  Create a SAX parser object. Then, according to what we were told on
  //  the command line, set the options.
  //
  SAXParser* parser = new SAXParser;
  parser->setValidationScheme(valScheme);
  parser->setDoNamespaces(doNamespaces);
  parser->setDoSchema(doSchema);
  parser->setValidationSchemaFullChecking(schemaFullChecking);
  

  //
  //  Create our SAX handler object and install it on the parser, as the
  //  document and error handler. We are responsible for cleaning them
  //  up, but since its just stack based here, there's nothing special
  //  to do.
  //
  ViewCellsParseHandlers handler(tree, viewCells, objects);
  parser->setDocumentHandler(&handler);
  parser->setErrorHandler(&handler);
  
  unsigned long duration;
  int errorCount = 0;
  // create a faux scope so that 'src' destructor is called before
  // XMLPlatformUtils::Terminate
  {
    //
    //  Kick off the parse and catch any exceptions. Create a standard
    //  input input source and tell the parser to parse from that.
    //
    //    StdInInputSource src;
    try
      {
	const unsigned long startMillis = XMLPlatformUtils::getCurrentMillis();
	parser->parse(filename.c_str());
	const unsigned long endMillis = XMLPlatformUtils::getCurrentMillis();
	duration = endMillis - startMillis;
	errorCount = parser->getErrorCount();
      }
    catch (const OutOfMemoryException&)
      {
	XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
	errorCount = 2;
	return false;
      }
    catch (const XMLException& e)
      {
	XERCES_STD_QUALIFIER cerr << "\nError during parsing: \n"
				  << StrX(e.getMessage())
				  << "\n" << XERCES_STD_QUALIFIER endl;
	errorCount = 1;
	return false;
      }

    
    // Print out the stats that we collected and time taken
    if (!errorCount) {
      XERCES_STD_QUALIFIER cout << filename << ": " << duration << " ms ("
				<< handler.GetElementCount() << " elems, "
				<< handler.GetAttrCount() << " attrs, "
				<< handler.GetSpaceCount() << " spaces, "
				<< handler.GetCharacterCount() << " chars)" << XERCES_STD_QUALIFIER endl;
    }
  }
  
  //
  //  Delete the parser itself.  Must be done prior to calling Terminate, below.
  //
  delete parser;
  
  XMLPlatformUtils::Terminate();
  
  if (errorCount > 0)
    return false;
  else
    return true;
}